﻿B4J=true
Group=Default Group
ModulesStructureVersion=1
Type=Class
Version=6.01
@EndOfDesignText@
#Event: BitmapReady (bmp As B4XBitmap)
'version 4.30
Sub Class_Globals
	Private mBuffer() As Byte
	Public mWidth, mHeight As Int
	Private ai, ri, gi, bi As Int
	Type PremultipliedColor (a As Int, r As Int, g As Int, b As Int)
	Type ARGBColor (a As Int, r As Int, g As Int, b As Int)
	Type DrawTask (Source As Object, SrcRect As B4XRect, TargetX As Int, TargetY As Int, _
		SkipBlending As Boolean, SrcScaleX As Float, SrcScaleY As Float, Degrees As Int, _
		Transform As Boolean, IsCompressedSource As Boolean, TargetBC As BitmapCreator)
	Type CompressedBC (Rows As List, mBuffer() As Byte, mWidth As Int, mHeight As Int, TargetRect As B4XRect, Cache As InternalCompressedBCCache)
	Type InternalCompressedBCCache (ColorsMap As Map, mBuffer() As Byte)
	#if B4J or B4A
	Type InternalAntiAliasingBuffer (IntsArray() As Int)
	#else
	Type InternalAntiAliasingBuffer (IntsArray() As Byte)
	#end if
	Private tempARGB As ARGBColor
	Private tempPM, tempPM2 As PremultipliedColor
#if B4A
	Private b4abmp As Bitmap
	Private jme As JavaObject
#else if B4J
	Private WriteableImage As JavaObject
	Private jme As JavaObject
#Else
	Private jme As NativeObject
	Private ShouldClearBeforeExtracting As Boolean
#end if
	Public TargetRect As B4XRect
	Private UnsignedFF As Int
	Private xui As XUI
	Private const CACHE_TRANSPARENT = 0, CACHE_SOLID = 1 As Byte
#if B4i
	Public const SAME_COLOR_LENGTH_FOR_CACHE As Int = 40
#else
	Public const SAME_COLOR_LENGTH_FOR_CACHE As Int = 20
#end if
	Public const MAX_SAME_COLOR_SIZE As Int = 5000
	#if Stats
	Public ClockForAsync As Long
	#End If
	Private SingleRow() As Byte
	Private columns As List
	Public AntiAliasDrawings As Boolean = True
End Sub

'Initializes the object and sets the bitmap dimensions.
'Note that XUI library must be referenced as well.
Public Sub Initialize (Width As Int, Height As Int)
	Dim b As Byte = 0xff
	UnsignedFF = b
	mWidth = Width
	mHeight = Height
	Dim mBuffer(mWidth * mHeight * 4) As Byte
	#if B4A or B4i
	ai = 3
	ri = 0
	gi = 1
	bi = 2
	#else if B4J
	ai = 3
	ri = 2
	gi = 1
	bi = 0
	
	#end if
	jme = Me
	TargetRect.Initialize(0, 0, mWidth, mHeight)
End Sub

'Gets the internal buffer.
Public Sub getBuffer() As Byte()
	Return mBuffer
End Sub

'Extracts the pixels from the SourceBitmap and copies them.
'Note that the source bitmap is first resized, if needed, to match the BitmapCreator dimensions.
Public Sub CopyPixelsFromBitmap (SourceBitmap As B4XBitmap)
	If SourceBitmap.Width <> mWidth Or SourceBitmap.Height <> mHeight Then
'		Log("resize: " & SourceBitmap.Scale)
		SourceBitmap = SourceBitmap.Resize(mWidth, mHeight, False)
	End If
	#if B4A
	Dim jo As JavaObject = SourceBitmap
	Dim format As Object = jo.RunMethod("getConfig", Null)
	If "ARGB_8888" <> format Then
		Log("Unsupported format: " & format)
		Return
	End If
	Dim bf As JavaObject
	bf = bf.InitializeStatic("java.nio.ByteBuffer").RunMethodJO("wrap", Array(mBuffer))
	jo.RunMethod("copyPixelsToBuffer", Array(bf))
	#else if B4J
	Dim jo As JavaObject = SourceBitmap
	Dim reader As JavaObject = jo.RunMethod("getPixelReader", Null)
	Dim PixelFormat As JavaObject
	PixelFormat.InitializeStatic("javafx.scene.image.PixelFormat")
	reader.RunMethod("getPixels", Array(0, 0, mWidth, mHeight, PixelFormat.RunMethod("getByteBgraPreInstance", Null), _
       mBuffer, 0, mWidth * 4))
	 #Else If B4i
	Dim no As NativeObject = Me
	no.RunMethod("UIImageToBuffer:::::", Array(SourceBitmap, no.ArrayToNSData(mBuffer), mWidth, mHeight, ShouldClearBeforeExtracting))
	ShouldClearBeforeExtracting = True
	#End If
End Sub

'Sets the color of the given pixel.
Public Sub SetColor(x As Int, y As Int, Clr As Int)
	ColorToARGB(Clr, tempARGB)
	SetARGB(x, y, tempARGB)
End Sub

'Converts color int to ARGBColor object.
'The Result parameter will hold the output.
Public Sub ColorToARGB (Clr As Int, Result As ARGBColor) As ARGBColor
	Result.a = Bit.And(0xff, Bit.ShiftRight(Clr, 24))
	Result.r = Bit.And(0xff, Bit.ShiftRight(Clr, 16))
	Result.g =  Bit.And(0xff, Bit.ShiftRight(Clr, 8))
	Result.b =  Bit.And(0xff, Clr)
	Return Result
End Sub

'Converts an ARGB color to PremultipliedColor.
'The Result parameter will hold the output.
Public Sub ARGBToPremultipliedColor (ARGB As ARGBColor, PM As PremultipliedColor) As PremultipliedColor
	Dim a As Float = ARGB.a / 255
	PM.a = ARGB.a
	PM.r = ARGB.r * a
	PM.g = ARGB.g * a
	PM.b = ARGB.b * a
	Return PM
End Sub

'Sets the color of the specified point.
Public Sub SetHSV(x As Int, y As Int, alpha As Int, h As Int, s As Float, v As Float)
	Dim r, g, b As Float
	Dim hi As Int = Floor(h / 60)
	Dim hbucket As Int =  hi Mod 6
	Dim f As Float = h / 60 - hi
	Dim p As Float = v * (1 - s)
	Dim q As Float = v  * (1 - f * s)
	Dim t As Float = v * (1 - (1 - f) * s)
	Select hbucket
		Case 0
			r = v
			g = t
			b = p
		Case 1
			r = q
			g = v
			b = p
		Case 2
			r = p
			g = v
			b = t
		Case 3
			r = p
			g = q
			b = v
		Case 4
			r = t
			g = p
			b = v
		Case 5
			r = v
			g = p
			b = q
	End Select
	tempARGB.a = alpha
	tempARGB.r = r * 255
	tempARGB.g = g * 255
	tempARGB.b = b * 255
	SetARGB(x, y, tempARGB)
End Sub

'Sets the color of the specified point.
Public Sub SetARGB(x As Int, y As Int, ARGB As ARGBColor)
	SetPremultipliedColor(x, y, ARGBToPremultipliedColor(ARGB, tempPM))
End Sub

'Sets the color of the specified point. 
Public Sub SetPremultipliedColor (x As Int, y As Int, Premultiplied As PremultipliedColor)
	Dim cp As Int = x * 4 + y * mWidth * 4
	#if B4i
	Bit.FastArraySetByte(mBuffer, cp + ri, Premultiplied.r)
	Bit.FastArraySetByte(mBuffer, cp + gi, Premultiplied.g)
	Bit.FastArraySetByte(mBuffer, cp + bi, Premultiplied.b)
	Bit.FastArraySetByte(mBuffer, cp + ai, Premultiplied.a)
	#else
	mBuffer(cp + ri) = Premultiplied.r
	mBuffer(cp + gi) = Premultiplied.g
	mBuffer(cp + bi) = Premultiplied.b
	mBuffer(cp + ai) = Premultiplied.a
	#End If
	
End Sub

'Gets the color of the given point as a premultiplied color. No conversion is required for this format.
'The Result parameter stores the output.
Public Sub GetPremultipliedColor (x As Int, y As Int, Result As PremultipliedColor) As PremultipliedColor
	Dim cp As Int = x * 4 + y * mWidth * 4
	#if B4i
	Result.r = Bit.FastArrayGetByte(mBuffer, cp + ri)
	Result.g = Bit.FastArrayGetByte(mBuffer, cp + gi)
	Result.b = Bit.FastArrayGetByte(mBuffer, cp + bi)
	Result.a = Bit.FastArrayGetByte(mBuffer, cp + ai)
	#else
	Result.r = mBuffer(cp + ri)
	Result.g = mBuffer(cp + gi)
	Result.b = mBuffer(cp + bi)
	Result.a = mBuffer(cp + ai)
	#End If
	Return Result
End Sub

'Gets the color of the given point as an int value.
Public Sub GetColor (x As Int, y As Int) As Int
	GetARGB(x, y, tempARGB)
	Return Bit.ShiftLeft(tempARGB.a, 24) + Bit.ShiftLeft(tempARGB.r, 16) + Bit.ShiftLeft(tempARGB.g, 8) + tempARGB.b
End Sub

'Gets the color of the given point as an ARGB color.
'The Result parameter stores the output.
Public Sub GetARGB (x As Int, y As Int, Result As ARGBColor) As ARGBColor
	Dim cp As Int = x * 4 + y * mWidth * 4
#if B4i
	Result.a = Bit.FastArrayGetByte(mBuffer, cp + ai)
	Dim af As Float = Result.a / 255
	Result.r = Bit.FastArrayGetByte(mBuffer, cp + ri) / af
	Result.g = Bit.FastArrayGetByte(mBuffer, cp + gi) / af
	Result.b = Bit.FastArrayGetByte(mBuffer, cp + bi) / af
#Else
	Result.a = Bit.And(0xff, mBuffer(cp + ai))
	Dim af As Float = Result.a / 255
	Result.r = Bit.And(0xff, mBuffer(cp + ri)) / af
	Result.g = Bit.And(0xff, mBuffer(cp + gi)) / af
	Result.b = Bit.And(0xff, mBuffer(cp + bi)) / af
#End If
	
	Return Result
End Sub

'Tests whether a given point is completely transparent.
Public Sub IsTransparent(x As Int, y As Int) As Boolean
	Dim cp As Int = x * 4 + y * mWidth * 4
	Return mBuffer(cp + ai) = 0
End Sub

'Converts the bytes buffer to a bitmap.
Public Sub getBitmap As B4XBitmap
	#if Stats
	Dim Clock As Long = DateTime.Now
	#End If
	#if B4A
	Dim bf As JavaObject
	bf = bf.InitializeStatic("java.nio.ByteBuffer").RunMethodJO("wrap", Array(mBuffer))
	If b4abmp.IsInitialized = False Then
		b4abmp.InitializeMutable(mWidth, mHeight)
	End If
	Dim jo As JavaObject = b4abmp
	jo.RunMethod("setDensity", Array(160))
	jo.RunMethod("copyPixelsFromBuffer", Array(bf))
	jo.RunMethod("prepareToDraw", Null)
	#if Stats
	Log($"Get Bitmap: ${DateTime.Now - Clock} ms"$)
	#End If
	Return b4abmp
	#else if B4J
	If WriteableImage.IsInitialized = False Then
		WriteableImage.InitializeNewInstance("javafx.scene.image.WritableImage", Array(mWidth, mHeight))
	End If
	Dim writer As JavaObject = WriteableImage.RunMethod("getPixelWriter", Null)
	Dim PixelFormat As JavaObject
	PixelFormat.InitializeStatic("javafx.scene.image.PixelFormat")
	writer.RunMethod("setPixels", Array(0, 0, mWidth, mHeight, PixelFormat.RunMethod("getByteBgraPreInstance", Null), _
		mBuffer, 0, mWidth * 4))
	#if Stats
	Log($"Get Bitmap: ${DateTime.Now - Clock} ms"$)
	#End If
	Return WriteableImage
	#Else
	Dim no As NativeObject = Me
	Dim bmp As Bitmap = no.RunMethod("bufferToUIImage:::", Array(no.ArrayToNSData(mBuffer), mWidth, mHeight))
	Return bmp
	#end if
End Sub



#Region ************  Drawing methods ***************
'Draws a line between (X1, Y1) to (X2, Y2).
'Even StrokeWidth values are decreased by one.
Public Sub DrawLine (X1 As Float, Y1 As Float, X2 As Float, Y2 As Float, Clr As Int, StrokeWidth As Int)
	If X1 = X2 And Y1 = Y2 Then Return
	If X2 < X1 Then
		Dim a As Float = X1
		X1 = X2
		X2 = a
		a = Y1
		Y1 = Y2
		Y2 = a
	End If
	Dim CheckBounds As Boolean
	Dim HalfWidth As Int = Ceil(StrokeWidth / 2)
	If X1 - HalfWidth < 0 Or X2 + HalfWidth >= mWidth Or Y1 - HalfWidth < 0 Or Y2 - HalfWidth < 0 _
			Or Y1 + HalfWidth >= mHeight - 1 Or Y2 + HalfWidth >= mHeight - 1 Then
		If X1 >= mWidth Or X2 < 0 Then Return
		If (Y1 >= mHeight And Y2 >= mHeight) Or (Y1 < 0 And Y2 < 0) Then Return
		CheckBounds = True
	End If
	If StrokeWidth = 1 Then
		DrawBresenhamLine(X1, Y1, X2, Y2, Clr, CheckBounds)
	Else
		DrawThickLine(X1, Y1, X2, Y2, Clr, CheckBounds, StrokeWidth)
	End If
End Sub

Private Sub DrawBresenhamLine(X0 As Float, Y0 As Float, X1 As Float, Y1 As Float, Clr As Int, CheckBounds As Boolean)
	ARGBToPremultipliedColor(ColorToARGB(Clr, tempARGB), tempPM)
	Dim sy, e2, err As Int
	Dim dx As Int = Abs(X1 - X0)
	Dim dy As Int = -Abs(Y1 - Y0)
	If Y0 < Y1 Then sy = 1 Else sy = -1
	Dim sybuffer As Int = mWidth * 4 * sy
	err = dx + dy
	Dim cp As Int = X0 * 4 + Y0 * mWidth * 4
	Do While True
		If CheckBounds = False Or (X0 >= 0 And X0 < mWidth And Y0 >= 0 And Y0 < mHeight) Then
			#if B4i
			Bit.FastArraySetByte(mBuffer, cp + ri, tempPM.r)
			Bit.FastArraySetByte(mBuffer, cp + gi, tempPM.g)
			Bit.FastArraySetByte(mBuffer, cp + bi, tempPM.b)
			Bit.FastArraySetByte(mBuffer, cp + ai, tempPM.a)
			#else
			mBuffer(cp + ri) = tempPM.r
			mBuffer(cp + gi) = tempPM.g
			mBuffer(cp + bi) = tempPM.b
			mBuffer(cp + ai) = tempPM.a
			#End If
		End If
		e2 = err + err
		If e2 >= dy Then
			If X0 = X1 Then Exit
			X0 = X0 + 1
			cp = cp + 4
			err = err + dy
		End If
		If e2 <= dx Then
			If Y0 = Y1 Then Exit
			Y0 = Y0 + sy
			cp = cp + sybuffer
			err = err + dx
		End If
	Loop
End Sub

Private Sub DrawThickLine(X0 As Float, Y0 As Float, X1 As Float, Y1 As Float, Clr As Int, CheckBounds As Boolean, StrokeWidth As Int)
	Dim TotalDX As Float = X1 - X0
	Dim TotalDY As Float = Y1 - Y0
	StrokeWidth = StrokeWidth - 1 + (StrokeWidth Mod 2)
	Dim halfwidth As Float = (StrokeWidth - 1)/ 2
	Dim alpha As Float = ATan2(Abs(TotalDY), TotalDX)
	Dim offsetx, offsety As Float
	offsetx = halfwidth * Sin(alpha)
	offsety = halfwidth * Cos(alpha)
	If Y1 < Y0 Then offsety = -offsety
	Dim MaxRowLength As Int = Min(StrokeWidth + TotalDX, Round(StrokeWidth / Sin(alpha)))
	Dim pm As PremultipliedColor = InitSingleRowForDrawing(Min(mWidth, MaxRowLength + 4), 2, Clr)
	#if B4i
	Dim columnleft() As Byte = columns.Get(0)
	Dim columnright() As Byte = columns.Get(1)
#else
	Dim columnleft() As Int = columns.Get(0)
	Dim columnright() As Int = columns.Get(1)
#End If
	If CheckBounds Then
		Dim oy As Int = Ceil(Abs(offsety))
		If X0 - Ceil(offsetx) >= 0 And X1 + Ceil(offsetx) < mWidth And Y0 - oy >= 0 And Y1 - oy >= 0 And _
			Y0 + oy < mHeight - 1 And Y1 + oy < mHeight - 1 Then
			CheckBounds = False
		End If
	End If
	Dim MinY As Int = Min(mHeight - 1, Max(0, Min(Y0 - offsety,Y1 + offsety)))
	Dim Maxy As Int = Min(mHeight - 1, Max(0, Max(Y0 - offsety,Y1 + offsety)))
	#if B4i
	Bit.FastArraySetInt(columnright, MinY, 0)
	Bit.FastArraySetInt(columnleft, MinY, 0)
	#Else
	columnright(MinY) = 0
	columnright(Maxy) = 0
	#End If
	DrawWuLine(Floor(X0 - offsetx), Round(Y0 + offsety), Floor(X1 - offsetx), Round(Y1 + offsety), CheckBounds, True, columnleft, pm)
	DrawWuLine(Ceil(X0 + offsetx), Round(Y0 - offsety), Ceil(X1 + offsetx), Round(Y1 - offsety), CheckBounds, False, columnright, pm)
	DrawWuLine(Floor(X0 - offsetx), Round(Y0 + offsety), Ceil(X0 + offsetx), Round(Y0 - offsety), CheckBounds, True, columnleft, pm)
	DrawWuLine(Floor(X1 - offsetx), Round(Y1 + offsety), Ceil(X1 + offsetx), Round(Y1 - offsety), CheckBounds, False, columnright, pm)
	For y = MinY To Maxy
		#if B4i
		Dim xstart As Int = Bit.FastArrayGetInt(columnleft, y) + 1
		Dim xend As Int = Bit.FastArrayGetInt(columnright, y)
		#else
		Dim xstart As Int = columnleft(y) + 1
		Dim xend As Int = columnright(y)
		#End If
		If xend > xstart Then
			Dim cp As Int = xstart * 4 + y * mWidth * 4
			Bit.ArrayCopy(SingleRow, 0, mBuffer, cp, (xend - xstart) * 4)
		End If
	Next
End Sub

Private Sub InitSingleRowForDrawing (Length As Int, NumberOfColumns As Int, Clr As Int) As PremultipliedColor
	Dim a As ARGBColor
	ColorToARGB(Clr, a)
	ARGBToPremultipliedColor(a, tempPM2)
	Dim pm As PremultipliedColor
	ARGBToPremultipliedColor(a, pm)
	If SingleRow.Length = 0 Then
		Dim SingleRow(mWidth * 4 + 4 * 4) As Byte
		columns.Initialize
	End If
	If columns.Size < NumberOfColumns Then
		For i = 0 To NumberOfColumns - 1
		#if B4i
			Dim col1(mHeight * 4) As Byte
		#else
			Dim col1(mHeight) As Int
		#End If
			columns.Add(col1)
		Next
	End If
	For i = 0 To Length * 4 - 1 Step 4
		#if B4i
		Bit.FastArraySetByte(SingleRow, i + ri, tempPM2.r)
		Bit.FastArraySetByte(SingleRow, i + gi, tempPM2.g)
		Bit.FastArraySetByte(SingleRow, i + bi, tempPM2.b)
		Bit.FastArraySetByte(SingleRow, i + ai, tempPM2.a)
		#else
		SingleRow(i + ri) = tempPM2.r
		SingleRow(i + gi) = tempPM2.g
		SingleRow(i + bi) = tempPM2.b
		SingleRow(i + ai) = tempPM2.a
		#End If
	Next
	Return pm
End Sub

#if B4i
Private Sub DrawWuLine (X0 As Int, Y0 As Int, X1 As Int, Y1 As Int, CheckBounds As Boolean, LeftSide As Boolean, column() As Byte, pmclr As PremultipliedColor)
#else
Private Sub DrawWuLine (X0 As Int, Y0 As Int, X1 As Int, Y1 As Int, CheckBounds As Boolean, LeftSide As Boolean, column() As Int, pmclr As PremultipliedColor)
#end if
	Dim steep As Boolean =  Abs(Y1 - Y0) > Abs(X1 - X0)
	Dim t As Int
	If steep Then
		t = X0
		X0 = Y0
		Y0 = t
		t = X1
		X1 = Y1
		Y1 = t
	End If
	If X0 > X1 Then
		t = X0
		X0 = X1
		X1 = t
		t = Y0
		Y0 = Y1
		Y1 = t
	End If
	Dim dy As Float = Y1 - Y0
	Dim dx As Float = X1 - X0
	Dim SignY As Int
	If dy > 0 Then SignY = 1 Else SignY = -1
	Dim gradient As Float
	If dx = 0 Then gradient = 1 Else gradient = dy / dx
	Dim x As Float
	Dim ix, iy As Int
	Dim fp As Float
	Dim intery As Float = Y0
	Dim t As Int
	If steep Then
		For iy = X0 To X1
			ix = Floor(intery)
			If CheckBounds = False Or (intery >= 0 And intery + 1 < mWidth And iy >= 0 And iy < mHeight) Then
				fp = intery - ix
				If LeftSide Then
					If AntiAliasDrawings Then DrawAntiAliasBorder(ix, iy, 1 - fp, pmclr, False)
					t = Max(0, Min(mWidth - 1, ix))
				Else
					t = Max(0, Min(mWidth - 1, ix + 1))
					If AntiAliasDrawings Then DrawAntiAliasBorder(ix + 1, iy, fp, pmclr, False)
				End If
				#if B4i
				Bit.FastArraySetInt(column, iy, t)
				#else
				column(iy) = t
				#End If
			Else if iy >= 0 And iy < mHeight Then
				t =  Max(0, Min(mWidth - 1, ix))
				#if B4i
				Bit.FastArraySetInt(column, iy, t)
				#else
				column(iy) = t
				#End If
			End If
			intery = intery + gradient
		Next
	Else
		x = X0
		If SignY > 0 Then
			Dim g As Float = 1 / gradient
			For iy = Y0 To Y1
				If CheckBounds Then
					If iy >= 0 And iy < mHeight Then
						t = Max(0, Min(mWidth - 1, x + 0.5))
						#if B4i
						Bit.FastArraySetInt(column, iy, t)
						#else
						column(iy) = t
						#End If
					End If
				Else
					t = x + 0.5
					#if B4i
					Bit.FastArraySetInt(column, iy, t)
					#else
					column(iy) = t
					#End If
				End If
				x = x + g
			Next
		Else
			Dim g As Float = -1 / gradient
			For iy = Y0 To Y1  Step -1
				If CheckBounds Then
					If iy >= 0 And iy < mHeight Then
						t = Max(0, Min(mWidth - 1, x + 0.5))
						#if B4i
						Bit.FastArraySetInt(column, iy, t)
						#else
						column(iy) = t
						#End If
					End If
				Else
					t = x + 0.5
					#if B4i
					Bit.FastArraySetInt(column, iy, t)
					#else
					column(iy) = t
					#End If
				End If
				x = x + g
			Next
		End If
		If AntiAliasDrawings Then
			For ix = X0 To X1
				iy = Floor(intery)
				If CheckBounds = False Or (ix >= 0 And ix < mWidth And iy >= 0 And iy + 1 < mHeight) Then
					fp = intery - iy
					#if B4i
					Dim ciy As Int = Bit.FastArrayGetInt(column, iy)
					Dim ciy1 As Int = Bit.FastArrayGetInt(column, iy + 1)
					#else
					Dim ciy As Int = column(iy)
					Dim ciy1 As Int = column(iy + 1)
					#End If
					If (LeftSide And ix <= ciy) Or (LeftSide = False And ix >= ciy) Then
						DrawAntiAliasBorder(ix, iy, 1 - fp, pmclr, False)
					End If
					If (LeftSide And ix <= ciy1) Or (LeftSide = False And ix >= ciy1) Then
						DrawAntiAliasBorder(ix, iy + 1, fp, pmclr, False)
					End If
				End If
				intery = intery + gradient
			Next
		End If
	End If
End Sub

'Draws a a circle.
Public Sub DrawCircle (X As Int, Y As Int, Radius As Int, Clr As Int, Filled As Boolean, StrokeWidth As Int)
	If x + Radius < 0 Or x - Radius >= mWidth Or y + Radius < 0 Or y - Radius >= mHeight Then Return
	Dim CheckBounds As Boolean = x - Radius < 0 Or x + Radius >= mWidth Or y - Radius < 0 Or y + Radius >= mHeight
	If Filled Or StrokeWidth >= Radius Then
		Dim pm As PremultipliedColor = InitSingleRowForDrawing(2 * Radius + 1, 1, Clr)
		#if B4i
		Dim col() As Byte = columns.Get(0)
		#else
		Dim col() As Int = columns.Get(0)
		#End If
		DrawWuCircle(X, Y, Radius - 1, CheckBounds, False, col, pm)
		For row = Max(0, Y - Radius + 1) To Min(mHeight, y + Radius) - 1
			#if B4i
			Dim dx As Int = Bit.FastArrayGetInt(col, row)
			#else
			Dim dx As Int = col(row)
			#end if
			If dx > 0 Then
				Dim X1 As Int = Max(0, x - dx)
				Dim x2 As Int = Min(mWidth - 1, x + dx)
				If x2 >= X1 Then
					Dim cp As Int = X1 * 4 + row * mWidth * 4
					Bit.ArrayCopy(SingleRow, 0, mBuffer, cp, (x2 - X1 + 1) * 4)
				End If
			End If
		Next
	Else
		Dim pm As PremultipliedColor = InitSingleRowForDrawing(2 * Radius + 1, 2, Clr)
		#if B4i
		Dim col1() As Byte = columns.Get(0)
		Dim col2() As Byte = columns.Get(1)
		#else
		Dim col1() As Int = columns.Get(0)
		Dim col2() As Int = columns.Get(1)
		#End If
		
		Dim SmallRadius As Int =  Radius - StrokeWidth
		DrawWuCircle(X, Y, Radius - 1, CheckBounds, False, col1, pm)
		DrawWuCircle(X, Y, SmallRadius, CheckBounds, True, col2, pm)
		For row = Max(0, Y - Radius + 1) To Min(mHeight, y + Radius) - 1
			#if B4i
			Dim dx As Int = Bit.FastArrayGetInt(col1, row)
			#else
			Dim dx As Int = col1(row)
			#end if
			If dx > 0 Then
				Dim X1 As Int = Max(0, x - dx)
				Dim x2 As Int = Min(mWidth - 1, x + dx)
				If x2 >= X1 Then
					If Abs(row - y) <= SmallRadius Then
						#if B4i
						Dim dx2 As Int = Bit.FastArrayGetInt(col2, row)
						#else
						Dim dx2 As Int = col2(row)
						#end if
						Dim sx1 As Int = Min(mWidth, Max(0, x - dx2))
						Dim sx2 As Int = Max(0, Min(mWidth, x + dx2))
						If sx1 > X1 Then
							Dim cp As Int = X1 * 4 + row * mWidth * 4
							Bit.ArrayCopy(SingleRow, 0, mBuffer, cp, (sx1 - X1) * 4)
						End If
						If sx2 < x2 Then
							Dim cp As Int = (sx2 + 1) * 4 + row * mWidth * 4
							Bit.ArrayCopy(SingleRow, 0, mBuffer, cp, (x2 - sx2) * 4)
						End If
					Else
						Dim cp As Int = X1 * 4 + row * mWidth * 4
						Bit.ArrayCopy(SingleRow, 0, mBuffer, cp, (x2 - X1 + 1) * 4)
					End If
				End If
			End If
		Next
	End If
End Sub

#if B4i
Private Sub DrawWuCircle (X0 As Int, Y0 As Int, Radius As Int, CheckBounds As Boolean, Inner As Boolean, column() As Byte, pm As PremultipliedColor)
#else
Private Sub DrawWuCircle (X0 As Int, Y0 As Int, Radius As Int, CheckBounds As Boolean, Inner As Boolean, column() As Int, pm As PremultipliedColor)
#end if
	Dim s As Int = Sin(cPI / 4) * Radius + 0.5
	Dim c As Int = Cos(cPI / 4) * Radius + 0.5
	Dim r2 As Float = Radius * Radius
	Dim x, y As Float
	Dim ix, iy As Int
	For iy = 0 To Radius
		x = Sqrt(r2 - iy * iy)
		ix = x
		If CheckBounds = False Or (Y0 - iy >= 0 And Y0 - iy < mHeight) Then
			#if B4i
			Bit.FastArraySetInt(column, Y0 - iy, ix)
			#else
			column(Y0 - iy) = ix
			#End If
		End If
		If CheckBounds = False Or (Y0 + iy >= 0 And Y0 + iy < mHeight) Then
			#if B4i
			Bit.FastArraySetInt(column, Y0 + iy, ix)
			#else
			column(Y0 + iy) = ix
			#End If
		End If
		If iy <= s And AntiAliasDrawings Then
			Dim fp As Float = x - ix
			If Inner Then
				DrawAntiAliasBorder(X0 + ix, Y0 - iy, 1 - fp, pm, CheckBounds)
				DrawAntiAliasBorder(X0 - ix, Y0 - iy, 1 - fp, pm, CheckBounds)
				DrawAntiAliasBorder(X0 + ix, Y0 + iy, 1 - fp, pm, CheckBounds)
				DrawAntiAliasBorder(X0 - ix, Y0 + iy, 1 - fp, pm, CheckBounds)
			Else
				DrawAntiAliasBorder(X0 + ix + 1, Y0 - iy, fp, pm, CheckBounds)
				DrawAntiAliasBorder(X0 + ix + 1, Y0 + iy, fp, pm, CheckBounds)
				DrawAntiAliasBorder(X0 - ix - 1, Y0 + iy, fp, pm, CheckBounds)
				DrawAntiAliasBorder(X0 - ix - 1, Y0 - iy, fp, pm, CheckBounds)
			End If
		End If
	Next
	If AntiAliasDrawings Then
		For ix = c To 0 Step - 1
			y = Sqrt(r2 - ix * ix)
			iy = y
			Dim fp As Float = y - iy
			If Inner Or ix = 0 Then
				DrawAntiAliasBorder(X0 + ix, Y0 + iy, 1 - fp, pm, CheckBounds)
				DrawAntiAliasBorder(X0 + ix, Y0 - iy, 1 - fp, pm, CheckBounds)
				DrawAntiAliasBorder(X0 - ix, Y0 + iy, 1 - fp, pm, CheckBounds)
				DrawAntiAliasBorder(X0 - ix, Y0 - iy, 1 - fp, pm, CheckBounds)
			Else
				DrawAntiAliasBorder(X0 + ix, Y0 + iy + 1, fp, pm, CheckBounds)
				DrawAntiAliasBorder(X0 - ix, Y0 + iy + 1, fp, pm, CheckBounds)
				DrawAntiAliasBorder(X0 + ix, Y0 - iy - 1, fp, pm, CheckBounds)
				DrawAntiAliasBorder(X0 - ix, Y0 - iy - 1, fp, pm, CheckBounds)
			End If
		Next
	End If
End Sub

Private Sub DrawAntiAliasBorder(x As Int, y As Int, c As Float, pm As PremultipliedColor, CheckBounds As Boolean)
	If c < 0.1 Then Return
	If CheckBounds And (x < 0 Or x >= mWidth Or y < 0 Or y >= mHeight) Then 
		Return
	Else If c >= 0.9 Then
		BlendPixel2(pm, x, y)
	Else
		tempPM2.r = pm.r * c
		tempPM2.g =  pm.g * c
		tempPM2.b =  pm.b * c
		tempPM2.a =  pm.a * c
		BlendPixel2(tempPM2, x, y)
	End If
End Sub

Public Sub DrawRect (Rect As B4XRect, Color As Int, Filled As Boolean, StrokeWidth As Int)
	InitSingleRowForDrawing(Min(mWidth, Max(Rect.Width, StrokeWidth)), 0, Color)
	If Rect.Left >= mWidth Or Rect.Right < 0 Or Rect.Top >= mHeight Or Rect.Bottom < 0 Then Return
	If Filled Then
		Dim StartY As Int = Max(0, Rect.Top)
		Dim EndY As Int = Min(mHeight - 1, Rect.Bottom - 1)
		Dim StartX As Int = Max(0, Rect.Left)
		Dim EndX As Int = Min(mWidth - 1, Rect.Right - 1)
		Dim length As Int =(EndX - StartX + 1) * 4
		For y = StartY * mWidth * 4 To EndY * mWidth * 4 Step mWidth * 4
			Bit.ArrayCopy(SingleRow, 0, mBuffer, y + StartX * 4, length)
		Next
	Else
		Dim Top As Int = Rect.Top
		Dim Left As Int = Rect.Left
		Dim Right As Int = Rect.Right
		Dim Bottom As Int = Rect.Bottom
		Dim cp As Int
		Dim lfixed As Int = Max(0, Left)
		Dim rfixed As Int = Min(mWidth, Right)
		For y = Max(0, Top) To Min(mHeight, Top + StrokeWidth) - 1
			cp = lfixed * 4 + y * 4 * mWidth
			Bit.ArrayCopy(SingleRow, 0, mBuffer, cp, (rfixed - lfixed) * 4)
		Next
		For y = Max(0, Bottom - StrokeWidth) To Min(mHeight, Bottom) - 1
			cp = lfixed * 4 + y * 4 * mWidth
			Bit.ArrayCopy(SingleRow, 0, mBuffer, cp, (rfixed - lfixed) * 4)
		Next
		Dim CheckBounds As Boolean = Left < 0 Or Top < 0 Or Bottom > mHeight Or Right > mWidth
		If CheckBounds = False Then
			Dim length As Int = StrokeWidth * 4
			For y = Top + StrokeWidth To Bottom - StrokeWidth - 1
				cp = Left * 4 + y * mWidth * 4
				Bit.ArrayCopy(SingleRow, 0, mBuffer, cp, length)
				cp = (Right - StrokeWidth) * 4 + y * mWidth * 4
				Bit.ArrayCopy(SingleRow, 0, mBuffer, cp, length)
			Next
		Else
			Dim lengthleft As Int = Max(0, Min(StrokeWidth, Left + StrokeWidth))
			Dim lengthright As Int = Max(0, Min(StrokeWidth, mWidth + StrokeWidth - Right))
			For y = Max(0, Top + StrokeWidth) To Min(Bottom - StrokeWidth, mHeight) - 1
				If lengthleft > 0 Then
					cp = lfixed * 4 + y * mWidth * 4
					Bit.ArrayCopy(SingleRow, 0, mBuffer, cp, lengthleft * 4)
				End If
				If lengthright > 0 Then
					cp = (rfixed - lengthright) * 4 + y * mWidth * 4
				End If
				Bit.ArrayCopy(SingleRow, 0, mBuffer, cp, lengthright * 4)
			Next
		End If
	End If
End Sub

'DEPRECATED: use DrawRect instead.
Public Sub FillRect (Color As Int, Rect As B4XRect)
	DrawRect(Rect, Color, True, 0)
End Sub

'Replaces all semi-transparent pixels in the Rect with NewColor.
'This can be useful to remove antialiasing effects.
Public Sub ReplaceSemiTransparentPixels (NewColor As Int, Rect As B4XRect)
	Dim StartX As Int = Max(0, Rect.Left)
	Dim EndX As Int = Min(mWidth - 1, Rect.Right - 1)
	Dim StartY As Int = Max(0, Rect.Top)
	Dim EndY As Int = Min(mHeight - 1, Rect.Bottom - 1)
	If EndX < StartX Or EndY < StartY Then Return
	ColorToARGB(NewColor, tempARGB)
	ARGBToPremultipliedColor(tempARGB, tempPM)
	For y = StartY To EndY
		For x = StartX To EndX 
			Dim cp As Int = x * 4 + y * mWidth * 4
			Dim b As Byte = mBuffer(cp + ai)
			If b <> UnsignedFF And b <> 0 Then
				SetPremultipliedColor(x, y, tempPM)
			End If
		Next
	Next
End Sub

'Draws a bitmap to the buffer.
'SkipBlending - Whether non-opaque pixels in the source bitmap should blend with the background.
Public Sub DrawBitmap (Bmp As B4XBitmap, TargetRect1 As B4XRect, SkipBlending As Boolean)
	If Bmp.Width <> TargetRect1.Width Or Bmp.Height <> TargetRect1.Height Then
		Bmp = Bmp.Resize(TargetRect1.Width, TargetRect1.Height, False)
	End If
	Dim Src As BitmapCreator
	Src.Initialize(Bmp.Width, Bmp.Height)
	Src.CopyPixelsFromBitmap(Bmp)
	Dim SrcRect As B4XRect
	SrcRect.Initialize(0, 0, Src.mWidth, Src.mHeight)
	DrawBitmapCreator(Src, SrcRect, TargetRect1.Left, TargetRect1.Top, SkipBlending)
End Sub

'Creates a DrawTask object.
Public Sub CreateDrawTask (Source As Object, SrcRect As B4XRect, TargetX As Int, TargetY As Int, SkipBlending As Boolean) As DrawTask
	Dim dt As DrawTask
	dt.Source = Source
	dt.SrcRect = SrcRect
	dt.TargetX = TargetX
	dt.TargetY = TargetY
	dt.SkipBlending = SkipBlending
	dt.SrcScaleX = 1
	dt.SrcScaleY = 1
	dt.TargetBC = Null
	Return dt
End Sub
'Draws the image stored in the Source BitmapCreator to this BitmapCreator.
'Source - Source BitmapCreator
'SrcRect - Defines the region in the Source BC that will be copied.
'TargetX / TargetY - Top left point in the destination BC.
'SkipBlending - Whether non-opaque pixels in the source bitmap should blend with the background. 
Public Sub DrawBitmapCreator (Source As BitmapCreator, SrcRect As B4XRect, TargetX As Int, TargetY As Int, SkipBlending As Boolean)
	#if Stats
	Dim Clock As Long = DateTime.Now
	#End If
	If TargetY >= mHeight Or TargetX >= mWidth Then Return
	If TargetX + SrcRect.Width < 0 Or TargetY + SrcRect.Height < 0 Then Return
	Dim SrcTop As Int = SrcRect.Top
	Dim SrcLeft As Int = SrcRect.Left
	If TargetX < 0 Then
		SrcLeft = SrcLeft - TargetX
		TargetX = 0
	End If
	If TargetY < 0 Then
		SrcTop = SrcTop - TargetY
		TargetY = 0
	End If
	Dim SrcWidth As Int = SrcRect.Right - SrcLeft
	Dim SrcHeight As Int = SrcRect.Bottom - SrcTop
	TargetY = Max(0, TargetY)
	Dim DeltaX As Int = Min(mWidth - 1 - TargetX, Min(SrcWidth - 1, Source.mWidth - 1 - SrcLeft))
	Dim DeltaY As Int = Min(mHeight - 1 - TargetY, Min(SrcHeight - 1, Source.mHeight - 1 - SrcTop))
	
	If SkipBlending Then
		If TargetX = 0 And TargetY = 0 And SrcLeft = 0 And SrcTop = 0 And SrcWidth = Source.mWidth And SrcWidth = mWidth And _
			SrcHeight = Source.mHeight And SrcHeight = mHeight Then
			Bit.ArrayCopy(Source.mBuffer, 0, mBuffer, 0, mBuffer.Length)
		Else
			'Dim cp As Int = x * 4 + y * mWidth * 4
			For y = TargetY To TargetY + DeltaY
				Bit.ArrayCopy(Source.mBuffer, SrcLeft * 4 + (y - TargetY + SrcTop) * 4 * Source.mWidth, mBuffer, _
				TargetX * 4 + y * 4 * mWidth, (DeltaX + 1) * 4)
			Next
		End If
		
	Else
		For x = TargetX To TargetX + DeltaX
			For y = TargetY To TargetY + DeltaY
				#if B4i
				Dim cp1 As Int = (x - TargetX + SrcLeft) * 4 + (y - TargetY + SrcTop) * Source.mWidth * 4
				Dim cp2 As Int = x * 4 + y * mWidth * 4
				Dim a1 As Byte = Bit.FastArrayGetByte(Source.mBuffer, cp1 + ai)
				Dim r1 As Byte = Bit.FastArrayGetByte(Source.mBuffer, cp1 + ri)
				Dim g1 As Byte = Bit.FastArrayGetByte(Source.mBuffer, cp1 + gi)
				Dim b1 As Byte = Bit.FastArrayGetByte(Source.mBuffer, cp1 + bi)
				
				If a1 <> UnsignedFF Then
					If a1 = 0 Then Continue
					Dim a2 As Byte = Bit.FastArrayGetByte(mBuffer, cp2 + ai)
					Dim r2 As Byte = Bit.FastArrayGetByte(mBuffer, cp2 + ri)
					Dim g2 As Byte = Bit.FastArrayGetByte(mBuffer, cp2 + gi)
					Dim b2 As Byte = Bit.FastArrayGetByte(mBuffer, cp2 + bi)
					Dim SrcAF As Float = 1 - (a1 / 255)
					r1 =  r1 + SrcAF *  r2
					g1 =  g1 + SrcAF *  g2
					b1 =  b1 + SrcAF *  b2
					a1 =  a1 + SrcAF *  a2
				End If
				Bit.FastArraySetByte(mBuffer, cp2 + ri, r1)
				Bit.FastArraySetByte(mBuffer, cp2 + gi, g1)
				Bit.FastArraySetByte(mBuffer, cp2 + bi, b1)
				Bit.FastArraySetByte(mBuffer, cp2 + ai, a1)
				#else
				BlendPixel(Source, x - TargetX + SrcLeft, y - TargetY + SrcTop, x, y)
				
				#End If
			Next
		Next
	End If
	#if Stats
	Log($"DrawBitmapCreator: ${DateTime.Now - Clock} ms"$)
	#End If
End Sub

Public Sub FlipCompressedBitmap (Source As CompressedBC, Horizontal As Boolean, Vertical As Boolean) As CompressedBC
	If Horizontal = False And Vertical = False Then Return Source
	Dim NewCBC As CompressedBC
	NewCBC.Initialize
	NewCBC.Rows.Initialize
	NewCBC.mWidth = Source.mWidth
	NewCBC.mHeight = Source.mHeight
	NewCBC.TargetRect = Source.TargetRect
	NewCBC.Cache = Source.Cache
	If Horizontal = False Then
		NewCBC.mBuffer = Source.mBuffer
		For y = Source.Rows.Size - 1 To 0 Step -1
			NewCBC.Rows.Add(Source.Rows.Get(y))
		Next
	Else
		Dim NewBuffer(Source.mBuffer.Length) As Byte
		NewCBC.mBuffer = NewBuffer
		Dim SourceLast As Int = Source.mBuffer.Length - 4
		For i = 0 To NewBuffer.Length - 3 Step 4
			#if B4i
			Bit.FastArraySetByte(NewBuffer, i, Bit.FastArrayGetByte(Source.mBuffer, SourceLast - i))
			Bit.FastArraySetByte(NewBuffer, i + 1, Bit.FastArrayGetByte(Source.mBuffer, SourceLast - i + 1))
			Bit.FastArraySetByte(NewBuffer, i + 2, Bit.FastArrayGetByte(Source.mBuffer, SourceLast - i + 2))
			Bit.FastArraySetByte(NewBuffer, i + 3, Bit.FastArrayGetByte(Source.mBuffer, SourceLast - i + 3))
			#else
			NewBuffer(i) = Source.mBuffer(SourceLast - i)
			NewBuffer(i + 1) = Source.mBuffer(SourceLast - i + 1)
			NewBuffer(i + 2) = Source.mBuffer(SourceLast - i + 2)
			NewBuffer(i + 3) = Source.mBuffer(SourceLast - i + 3)
			#End If
			
		Next
		
		Dim index As Int
		For y = Source.Rows.Size - 1 To 0 Step -1
			Dim RowSize As Int
			#if B4i
			Dim Row() As Byte = Source.Rows.Get(y)
			RowSize = Row.Length / 4
			#else
			Dim row As List = Source.Rows.Get(y)
			RowSize = row.Size
			#End If
			#if B4i
			Dim NewRow(Row.Length) As Byte
			Bit.FastArraySetInt(NewRow, 0, index)
			#else
			Dim NewRow As List
			NewRow.Initialize
			NewRow.Add(index)
			#End If
			For i = RowSize - 2 To 1 Step -2
				#if B4i
				Dim state As Byte = Bit.FastArrayGetInt(Row, i)
				Dim count As Int = Bit.FastArrayGetInt(Row, i + 1)
				Bit.FastArraySetInt(NewRow, RowSize - 2 - i + 1, state)
				Bit.FastArraySetInt(NewRow, RowSize - 2 - i + 2, count)
				#else
				Dim state As Byte = row.Get(i)
				Dim count As Int = row.Get(i + 1)
				NewRow.Add(state)
				NewRow.Add(count)
				#End If
				If state = CACHE_SOLID Then
					index = index + count
				End If
			Next
			If Vertical Then
				NewCBC.Rows.Add(NewRow)
			Else
				NewCBC.Rows.InsertAt(0, NewRow)
			End If
		Next
	End If
	Return NewCBC
End Sub


'Private Sub PrintBCRow(index As Int)
'	Dim sb As StringBuilder
'	sb.Initialize
'	For x = 0 To mWidth - 1
'		'If x Mod 10 = 0 Then sb.Append(CRLF)
'		sb.Append(PrintColor(mBuffer, mWidth * index * 4 + x * 4)).Append(",")
'	Next
'	Log("******************* BC: " & index)
'	Log(sb)
'End Sub
'
'Private Sub PrintCBCRow(cbc As CompressedBC, y As Int)
'	Dim row As List = cbc.Rows.Get(y)
'	Dim sb As StringBuilder
'	sb.Initialize
'	Dim index As Int = row.Get(0)
'	Dim bc As ByteConverter
'	For i = 1 To row.Size - 1 Step 2
'		Dim state As Byte = row.Get(i)
'		Dim count As Int = row.Get(i + 1)
'		If state = CACHE_TRANSPARENT Then
'			For x = 0 To count - 1
'				sb.Append(bc.HexFromBytes(Array As Byte(0, 0, 0, 0))).Append(",")
'			Next
'		Else If state = CACHE_SOLID Then
'			For x = 0 To count - 1
'				sb.Append(PrintColor(cbc.mBuffer, (index + x) * 4)).Append(",")
'			Next
'			index = index + count
'		Else If state > CACHE_SOLID Then
'			For x = 0 To state - 1
'				sb.Append(bc.HexFromBytes(Array As Byte(0xbb, 0xbb, 0xbb, 0xbb))).Append(",")
'			Next
'		End If
'	Next
'	Log("******************* cbc: " & y)
'	Log(sb)
'End Sub
'
'Private Sub PrintColor(buffer() As Byte, index As Int) As String
'	Dim b(4) As Byte
'	b(0) = buffer(index + ai)
'	b(1) = buffer(index + ri)
'	b(2) = buffer(index + gi)
'	b(3) = buffer(index + bi)
'	Dim bc As ByteConverter
'	Return bc.HexFromBytes(b)
'End Sub

'Draws a CompressedBC object created with BitmapCreator.ExtractCompressBC.	
Public Sub DrawCompressedBitmap (Source As CompressedBC, SrcRect As B4XRect, TargetX As Int, TargetY As Int)
	#if Stats
	Dim Clock As Long = DateTime.Now
	#End If
	If TargetY >= mHeight Or TargetX >= mWidth Then Return
	If TargetX + SrcRect.Width < 0 Or TargetY + SrcRect.Height < 0 Then Return
	Dim SrcTop As Int = SrcRect.Top
	Dim SrcLeft As Int = SrcRect.Left
	If TargetX < 0 Then
		SrcLeft = SrcLeft - TargetX
		TargetX = 0
	End If
	If TargetY < 0 Then
		SrcTop = SrcTop - TargetY
		TargetY = 0
	End If
	Dim SrcWidth As Int = SrcRect.Right - SrcLeft
	Dim SrcHeight As Int = SrcRect.Bottom - SrcTop
	TargetY = Max(0, TargetY)
	Dim DeltaX As Int = Min(mWidth - 1 - TargetX, Min(SrcWidth - 1, Source.mWidth - 1 - SrcLeft))
	Dim DeltaY As Int = Min(mHeight - 1 - TargetY, Min(SrcHeight - 1, Source.mHeight - 1 - SrcTop))
	For y = TargetY To TargetY + DeltaY
		Dim RowSize As Int
		#if B4i
		Dim Row() As Byte = Source.Rows.Get(y - TargetY + SrcTop)
		RowSize = Row.Length / 4
		#else
		Dim row As List = Source.Rows.Get(y - TargetY + SrcTop)
		RowSize = row.Size
		#End If
		Dim x As Int = 0
		Dim FixedX, FixedCount As Int
		
		#if B4i
		Dim CBCDataOffset As Int = Bit.FastArrayGetInt(Row, 0)
		#else
		Dim CBCDataOffset As Int = row.Get(0)
		#End If
		For i = 1 To RowSize - 1 Step 2
			#if B4i
			Dim state As Byte = Bit.FastArrayGetInt(Row, i)
			Dim count As Int = Bit.FastArrayGetInt(Row, i + 1)
			#else
			Dim state As Byte = row.Get(i)
			Dim count As Int = row.Get(i + 1)
			#End If
			Dim cache_position As Int
			If state > CACHE_SOLID Then
				cache_position = count
				count = state
			End If
			
			If SrcLeft < x Then
				FixedX = x
				FixedCount = count
			Else
				If SrcLeft >= x + count Then
					x = x + count
					If state = CACHE_SOLID Then CBCDataOffset = CBCDataOffset + count
					If x > DeltaX + SrcLeft Then Exit
					Continue
				Else
					FixedX = SrcLeft
					FixedCount = count - (SrcLeft - x)
					If state  = CACHE_SOLID Then CBCDataOffset = CBCDataOffset + SrcLeft - x
				End If
			End If
			
			
			If state = CACHE_TRANSPARENT Then
			
			Else If state = CACHE_SOLID Then
				Bit.ArrayCopy(Source.mBuffer, CBCDataOffset * 4, mBuffer, _
						(TargetX + FixedX - SrcLeft) * 4 + y * 4 * mWidth, Min(FixedCount, (DeltaX + 1 - (x - SrcLeft))) * 4)
				CBCDataOffset = CBCDataOffset + FixedCount
			Else 'same color
				'cache_position
				Bit.ArrayCopy(Source.Cache.mBuffer, cache_position , mBuffer, _
						(TargetX + FixedX - SrcLeft) * 4 + y * 4 * mWidth, Min(FixedCount, (DeltaX + 1 - (x - SrcLeft))) * 4)
			End If
				
			x = x + count
			If x > DeltaX + SrcLeft Then Exit
		Next
	Next
	#if Stats
	Log($"DrawCompressedBitmap: ${DateTime.Now - Clock} ms"$)
	#End If
End Sub

'Rotates with antialiasing.
Public Sub DrawRotatedCBC (cbc As CompressedBC, Degrees As Float, Width As Int, Height As Int, AABuffer As InternalAntiAliasingBuffer)
	Dim srca As ARGBColor
	Dim cosl As Float = CosD(Degrees)
	Dim sinl As Float = SinD(Degrees)
	Dim srccx As Float = cbc.mWidth / 2
	Dim srccy As Float = cbc.mHeight / 2
	Dim targetcx As Float = Width / 2
	Dim targetcy As Float = Height / 2
	Dim TargetX, TargetY As Int
	#if B4i
	Dim dx(4 * 4), dy(4 * 4) As Byte
	Bit.FastArraySetFloat(dx, 1, 0.5 * cosl)
	Bit.FastArraySetFloat(dy, 1,  0.5 * sinl)
	Bit.FastArraySetFloat(dx, 2, -0.5 * sinl)
	Bit.FastArraySetFloat(dy, 2,  0.5 * cosl)
	Bit.FastArraySetFloat(dx, 3, cosl / 2 - sinl / 2)
	Bit.FastArraySetFloat(dy, 3,  sinl / 2 + cosl / 2)
	Dim PixelDeltaX As Float = 2 * Bit.FastArrayGetFloat(dx, 1)
	Dim PixelDeltaY As Float = 2 * Bit.FastArrayGetFloat(dy, 1)
	Dim IntsArray() As Byte = AABuffer.IntsArray
	#else
	Dim IntsArray() As Int = AABuffer.IntsArray
	Dim dx(4), dy(4) As Float
	dx(1) = 0.5 * cosl
	dy(1) = 0.5 * sinl
	dx(2) = -0.5 * sinl
	dy(2) = 0.5 * cosl
	dx(3) = cosl / 2 - sinl / 2
	dy(3) = sinl / 2 + cosl / 2
	Dim PixelDeltaX As Float = 2 * dx(1)
	Dim PixelDeltaY As Float = 2 * dy(1)
	#End If
	Dim RowLen As Int = Width * 5
	Dim index As Int
	'reset the counts
	For yy = 0 To Height - 1
		Dim index As Int = yy * RowLen
		For xx = 0 To Width - 1
			#if B4i
			Bit.FastArraySetInt(IntsArray, index + 4, 0)
			#else
			IntsArray(index + 4) = 0
			#End If
			index = index + 5
		Next
	Next
	#if B4i
	Dim StubRow(3 * 4) As Byte
	Bit.FastArraySetInt(StubRow, 2, cbc.mWidth)
	#Else
	Dim StubRow As List = Array(0, 0, cbc.mWidth)
	#End If
	Dim ystart As Float = (0 - srccx) * sinl + (0 - srccy) * cosl + targetcy
	For srcy = -1 To cbc.mHeight
		Dim y As Float = ystart
		Dim x As Float = (0 - srccx) * cosl - (srcy - srccy) * sinl + targetcx
		If srcy = -1 Or srcy = cbc.mHeight Then
			#if B4i
			Dim Row() As Byte = StubRow
			Dim RowSize As Int = 3
			#else
			Dim row As List = StubRow
			Dim RowSize As Int = row.Size
			#End If
		
		Else
			#if B4i
			Dim Row() As Byte = cbc.Rows.Get(srcy)
			Dim RowSize As Int = Row.Length / 4
			#else
			Dim row As List = cbc.Rows.Get(srcy)
			Dim RowSize As Int = row.size
			#End If
		End If
		
		#if B4i
		Dim CBCDataOffset As Int = Bit.FastArrayGetInt(Row, 0) * 4
		#else
		Dim CBCDataOffset As Int = row.Get(0) * 4
		#End If
		Dim StartI = 1, EndI = RowSize - 1 As Int
		#if B4i
		If Bit.FastArrayGetInt(Row, 0) <> CACHE_TRANSPARENT Then StartI = -1
		If Bit.FastArrayGetInt(Row, RowSize - 2) <> CACHE_TRANSPARENT Then EndI = RowSize + 1
		#else
		If row.Get(0) <> CACHE_TRANSPARENT Then StartI = -1
		If row.Get(row.Size - 2) <> CACHE_TRANSPARENT Then EndI = RowSize + 1
		#End If
		For i = StartI To EndI Step 2
			#if B4i
			If i = -1 Or i = RowSize Then
				Dim state As Byte = CACHE_TRANSPARENT
				Dim count As Int = 1
			Else
				Dim state As Byte = Bit.FastArrayGetInt(Row, i)
				Dim count As Int = Bit.FastArrayGetInt(Row, i + 1)
			End If
			
			#else
			If i = -1 Or i = RowSize Then
				Dim state As Byte = CACHE_TRANSPARENT
				Dim Count As Int = 1
			Else
				Dim state As Byte = row.Get(i)
				Dim Count As Int = row.Get(i + 1)
			End If
			#End If
			
			If state > CACHE_SOLID Then
				Dim cache_position As Int = Count
				Count = state
				#if B4i
				srca.a = Bit.FastArrayGetByte(cbc.Cache.mBuffer, cache_position + ai)
				Dim af As Float = srca.a / 255
				srca.r = Bit.FastArrayGetByte(cbc.Cache.mBuffer, cache_position + ri) / af
				srca.g = Bit.FastArrayGetByte(cbc.Cache.mBuffer, cache_position + gi) / af
				srca.b = Bit.FastArrayGetByte(cbc.Cache.mBuffer, cache_position + bi) / af
				#Else
				srca.a = Bit.And(0xff, cbc.Cache.mBuffer(cache_position + ai))
				Dim af As Float = srca.a / 255
				srca.r = Bit.And(0xff, cbc.Cache.mBuffer(cache_position + ri)) / af
				srca.g = Bit.And(0xff, cbc.Cache.mBuffer(cache_position + gi)) / af
				srca.b = Bit.And(0xff,cbc.Cache.mBuffer(cache_position + bi)) / af
				#end if
			Else If state = CACHE_TRANSPARENT Then
				srca.a = 0
				srca.r = 0
				srca.g = 0
				srca.b = 0
			End If
			For c = 0 To Count - 1
				If state = CACHE_SOLID Then
					#if B4i
					srca.a = Bit.FastArrayGetByte(cbc.mBuffer, CBCDataOffset + ai)
					Dim af As Float = srca.a / 255
					srca.r = Bit.FastArrayGetByte(cbc.mBuffer, CBCDataOffset + ri) / af
					srca.g = Bit.FastArrayGetByte(cbc.mBuffer, CBCDataOffset + gi) / af
					srca.b = Bit.FastArrayGetByte(cbc.mBuffer, CBCDataOffset + bi) / af
					#Else
					srca.a = Bit.And(0xff, cbc.mBuffer(CBCDataOffset + ai))
					Dim af As Float = srca.a / 255
					srca.r = Bit.And(0xff, cbc.mBuffer(CBCDataOffset + ri)) / af
					srca.g = Bit.And(0xff, cbc.mBuffer(CBCDataOffset + gi)) / af
					srca.b = Bit.And(0xff,cbc.mBuffer(CBCDataOffset + bi)) / af
					#end if
					CBCDataOffset = CBCDataOffset + 4
				End If
				For ii = 0 To 3
					#if B4i
					TargetX = x + Bit.FastArrayGetFloat(dx, ii) + 0.5
					TargetY = y + Bit.FastArrayGetFloat(dy, ii) + 0.5
					#else
					TargetX = x + dx(ii) + 0.5
					TargetY = y + dy(ii) + 0.5
					#End If
					If TargetX < 0 Or TargetY < 0 Or TargetX >= Width Or TargetY >= Height Then
						Continue
					End If
					Dim index As Int = TargetX * 5 + TargetY * RowLen
					#if B4i
					Dim PointCount As Int = Bit.FastArrayGetInt(IntsArray, index + 4)
					If PointCount = 0 Then
						Bit.FastArraySetInt(IntsArray, index, srca.a)
						Bit.FastArraySetInt(IntsArray, index + 1, srca.r)
						Bit.FastArraySetInt(IntsArray, index + 2, srca.g)
						Bit.FastArraySetInt(IntsArray, index + 3, srca.b)
					Else If srca.a > 0 Then
						Bit.FastArraySetInt(IntsArray, index, Bit.FastArrayGetInt(IntsArray, index) + srca.a)
						Bit.FastArraySetInt(IntsArray, index + 1, Bit.FastArrayGetInt(IntsArray, index + 1) + srca.r)
						Bit.FastArraySetInt(IntsArray, index + 2, Bit.FastArrayGetInt(IntsArray, index + 2) + srca.g)
						Bit.FastArraySetInt(IntsArray, index + 3, Bit.FastArrayGetInt(IntsArray, index + 3) + srca.b)
					End If
					Bit.FastArraySetInt(IntsArray, index + 4, PointCount + 1)
					#else
					Dim PointCount As Int = IntsArray(index + 4)
					If PointCount = 0 Then
						IntsArray(index) = srca.a
						IntsArray(index + 1) = srca.r
						IntsArray(index + 2) = srca.g
						IntsArray(index + 3) = srca.b
					Else If srca.a > 0 Then
						IntsArray(index) = IntsArray(index) + srca.a
						IntsArray(index + 1) = IntsArray(index + 1) + srca.r
						IntsArray(index + 2) = IntsArray(index + 2) + srca.g
						IntsArray(index + 3) = IntsArray(index + 3) + srca.b
					End If
					IntsArray(index + 4) = PointCount + 1
					#End If
					
				Next
				x = x + PixelDeltaX
				y = y + PixelDeltaY
			Next
		Next
		#if B4i
		ystart = ystart + Bit.FastArrayGetFloat(dy, 2) * 2
		#else
		ystart = ystart + dy(2) * 2
		#End If
	Next
	
	Dim a As ARGBColor
	For y = 0 To Height - 1
		Dim index As Int = y * RowLen
		#if B4i
		Dim cp As Int = y * mWidth * 4
		#End If
		For x = 0 To Width - 1
			#if B4i
			If Bit.FastArrayGetInt(IntsArray, index) > 0 Then
				Dim count As Int = Bit.FastArrayGetInt(IntsArray, index + 4)
				If count > 0 Then
					a.a = Bit.FastArrayGetInt(IntsArray, index) / count
					Dim af As Float = a.a / 255
					a.r = Bit.FastArrayGetInt(IntsArray, index + 1) / count * af
					a.g = Bit.FastArrayGetInt(IntsArray, index + 2) / count * af
					a.b = Bit.FastArrayGetInt(IntsArray, index + 3) / count * af
					Bit.FastArraySetByte(mBuffer, cp + ri, a.r)
					Bit.FastArraySetByte(mBuffer, cp + gi, a.g)
					Bit.FastArraySetByte(mBuffer, cp + bi, a.b)
					Bit.FastArraySetByte(mBuffer, cp + ai, a.a)
				End If
			End If
			cp = cp + 4
			#else
			If IntsArray(index) > 0 Then
				Dim Count As Int = IntsArray(index + 4)
				If Count > 0 Then
					a.a = IntsArray(index) / Count
					a.r = IntsArray(index + 1) / Count
					a.g = IntsArray(index + 2) / Count
					a.b = IntsArray(index + 3) / Count
					SetARGB(x, y, a)
				End If
			End If
			#End If
			
			index = index + 5
		Next
	Next
End Sub

Public Sub ExtractCompressedBC (Rect As B4XRect, Cache As InternalCompressedBCCache) As CompressedBC
	Dim cbc As CompressedBC
	cbc.Initialize
	cbc.Rows.Initialize
	cbc.mWidth = Rect.Width
	cbc.mHeight = Rect.Height
	cbc.TargetRect.Initialize(0, 0, Rect.Width, Rect.Height)
	cbc.Cache = Cache
	Dim Left = Rect.Left, Right = Rect.Right As Int
	Dim Top = Rect.Top, Bottom = Rect.Bottom As Int
	Dim Index As Int
	Dim LastA As Int, LastR, LastG, LastB As Byte
	Dim CheckForSameColor As Boolean = Cache.ColorsMap.Size < MAX_SAME_COLOR_SIZE
	Dim ColorKey As Int = 0
	Dim ColorPosition As Int = -1
	Dim SameColorCount As Int
	#if B4i
	Dim RowIndex As Int
	Dim Row(2000) As Byte
	#End If
	For y = Top To Bottom - 1
		#if B4i
		RowIndex = 0
		Bit.FastArraySetInt(Row, RowIndex, Index) : RowIndex = RowIndex + 1
		#else
		Dim Row As List
		Row.Initialize
		Row.Add(Index)
		#End If
		
		SameColorCount = 0
		LastA = -1000
		LastR = 0
		LastG = 0
		LastB = 0
		Dim EndOfLastSection As Int = Left - 1
		Dim CurrentState As Int
		For x = Left To Right - 1
			Dim offset As Int = x * 4 + y * mWidth * 4
			Dim NewState As Byte
			#if B4I
			Dim a As Int = Bit.FastArrayGetByte(mBuffer, offset + ai)
			#else
			Dim a As Int = mBuffer(offset + ai)
			#End If
			
			If a = UnsignedFF Then
				NewState = CACHE_SOLID
			Else If a = 0 Then
				NewState = CACHE_TRANSPARENT
			Else
				If Bit.And(0xff, a) > 30 Then
					NewState = CACHE_SOLID
				Else
					NewState = CACHE_TRANSPARENT
				End If
			End If
			If x = Left Then
				CurrentState = NewState
			End If
			If NewState = CACHE_SOLID And CheckForSameColor Then
				#if B4i
				Dim r As Byte = Bit.FastArrayGetByte(mBuffer, offset + ri)
				Dim g As Byte = Bit.FastArrayGetByte(mBuffer, offset + gi)
				Dim b As Byte = Bit.FastArrayGetByte(mBuffer, offset + bi)
				#else
				Dim r As Byte = mBuffer(offset + ri)
				Dim g As Byte = mBuffer(offset + gi)
				Dim b As Byte = mBuffer(offset + bi)
				#End If
				
				If a = LastA And r = LastR And g = LastG And b = LastB Then
					SameColorCount = SameColorCount + 1
				Else
					LastA = a : LastR = r : LastG = g : LastB = b
					SameColorCount = 1
				End If
				If SameColorCount = SAME_COLOR_LENGTH_FOR_CACHE  Then
					If x - EndOfLastSection > SAME_COLOR_LENGTH_FOR_CACHE Then
						Dim PrevCount As Int = x - EndOfLastSection - SAME_COLOR_LENGTH_FOR_CACHE
						#if B4i
						Bit.FastArraySetInt(Row, RowIndex, CACHE_SOLID) : RowIndex = RowIndex + 1
						Bit.FastArraySetInt(Row, RowIndex, PrevCount) : RowIndex = RowIndex + 1
						#else
						Row.Add(CACHE_SOLID)
						Row.Add(PrevCount)
						#End If
						Index = Index + PrevCount
					End If
					#if B4i
					Bit.FastArraySetInt(Row, RowIndex, SAME_COLOR_LENGTH_FOR_CACHE) : RowIndex = RowIndex + 1
					#else
					Row.Add(SAME_COLOR_LENGTH_FOR_CACHE)
					#End If
					Dim key As Int = LastA + Bit.ShiftLeft(Bit.And(0xff, LastR), 16) + Bit.ShiftLeft(Bit.And(0xff, LastG), 8) + Bit.And(0xff, LastB)
					If key <> ColorKey Then
						ColorPosition = Cache.ColorsMap.GetDefault(key, -1)
						If ColorPosition = -1 Then
							ColorPosition = Cache.ColorsMap.Size * SAME_COLOR_LENGTH_FOR_CACHE * 4
							Bit.ArrayCopy(mBuffer, offset  + 4 - SAME_COLOR_LENGTH_FOR_CACHE * 4, Cache.mBuffer, ColorPosition, SAME_COLOR_LENGTH_FOR_CACHE  * 4)
							Cache.ColorsMap.Put(key, ColorPosition)
						End If
						ColorKey = key
					End If
					#if B4i
					Bit.FastArraySetInt(Row, RowIndex, ColorPosition) : RowIndex = RowIndex + 1
					#else
					Row.Add(ColorPosition)
					#End If
					SameColorCount = 0
					CheckForSameColor = Cache.ColorsMap.Size < MAX_SAME_COLOR_SIZE
					EndOfLastSection = x
					CurrentState = CACHE_SOLID
					LastA = -1000
					Continue
				End If
			End If
			If NewState <> CurrentState Then
				'x is now the pixel with the new state!
				Dim LastSectionCount As Int = x - 1 - EndOfLastSection
				If LastSectionCount > 0 Then
					#if B4i
					Bit.FastArraySetInt(Row, RowIndex, CurrentState) : RowIndex = RowIndex + 1
					Bit.FastArraySetInt(Row, RowIndex, LastSectionCount) : RowIndex = RowIndex + 1
					#else
					Row.Add(CurrentState)
					Row.Add(LastSectionCount)
					#End If
					If CurrentState = CACHE_SOLID Then
						Index = Index + LastSectionCount
					End If
				End If
				EndOfLastSection = x - 1
				CurrentState = NewState
				If NewState = CACHE_SOLID Then SameColorCount = 1 Else SameColorCount = 0
			End If
		Next
		If EndOfLastSection < Right - 1 Then
			Dim LastSectionCount As Int = Right - 1 - EndOfLastSection
			#if B4i
			Bit.FastArraySetInt(Row, RowIndex, CurrentState) : RowIndex = RowIndex + 1
			Bit.FastArraySetInt(Row, RowIndex, LastSectionCount) : RowIndex = RowIndex + 1
			#else
			Row.Add(CurrentState)
			Row.Add(LastSectionCount)
			#End If
			If CurrentState = CACHE_SOLID  Then
				Index = Index + LastSectionCount
			End If
		End If
		#if B4i
		Dim brow(RowIndex * 4) As Byte
		Bit.ArrayCopy(Row, 0, brow, 0, RowIndex * 4)
		cbc.Rows.Add(brow)
		#else
		cbc.Rows.Add(Row)
		Dim c As Int
		For i = 1 To Row.Size - 1 Step 2
			Dim state As Int = Row.Get(i)
			If state > CACHE_SOLID Then
				c = c + state
			Else
				c = c + Row.Get(i + 1)
			End If
		Next
'		Log($"${y}: ${c}"$)
		#End If
	Next
	Dim data(Index * 4) As Byte
	cbc.mBuffer = data
	For y = 0 To cbc.Rows.Size - 1
		#if B4i
		Dim brow() As Byte = cbc.Rows.Get(y)
		Dim offsetInCBC As Int = Bit.FastArrayGetInt(brow, 0)
		Dim RowSize As Int = brow.Length / 4
		#else
		Dim Row As List = cbc.Rows.Get(y)
		Dim offsetInCBC As Int = Row.Get(0)
		Dim RowSize As Int = Row.Size
		#End If
		
		Dim RowOffet As Int = 0
		For i = 1 To RowSize - 1 Step 2
			#if B4i
			Dim state As Int = Bit.FastArrayGetInt(brow, i)
			Dim count As Int = Bit.FastArrayGetInt(brow, i + 1)
			#else
			Dim state As Int = Row.Get(i)
			Dim count As Int = Row.Get(i + 1)
			#end if	
			If state = CACHE_SOLID Then
				Bit.ArrayCopy(mBuffer, (Top + y) * mWidth * 4 + RowOffet * 4, data, offsetInCBC * 4, count * 4)
				offsetInCBC = offsetInCBC + count
			Else If state > CACHE_SOLID Then
				count = state
			End If
			RowOffet = RowOffet + count
		Next
	Next
'	PrintBCRow(0)
'	PrintCBCRow(cbc, 0)
	Return cbc
End Sub



'Asynchronously draws all the draw tasks. Note that no other drawings should be made until the Complete event is raised.
'Example: <code>bc1.DrawBitmapCreatorsAsync(Me, "BC", Tasks)
'Wait For BC_BitmapReady (bmp As B4XBitmap)
'</code>
Public Sub DrawBitmapCreatorsAsync(Target As Object, EventName As String, DrawTasks As List)
	#if Stats
	ClockForAsync = DateTime.Now
	#End If
	EventName = EventName.ToLowerCase
	If xui.IsB4i Then
		jme.RunMethod("DrawBitmapCreatorsAsync:::", Array(DrawTasks, EventName & "_bitmapready:", Target))
	Else
		jme.RunMethod("DrawBitmapCreatorsAsync", Array(DrawTasks, EventName, Target))
	End If
End Sub


'Draws a scaled and / or rotated BitmapCreator. 
'Task.TargetX / TargetY define the target center. This is different than the behavior in other methods.
'Task.SkipBlending = True will treat all pixels with alpha > 0 as solid colors and other pixels as transparent.
Public Sub DrawBitmapCreatorTransformed (Task As DrawTask)
	#if Stats
	Dim Clock As Long = DateTime.Now
	#End If
	Dim Degrees As Int = Task.Degrees
	Dim SrcScaleX As Float = Task.SrcScaleX
	Dim SrcScaleY As Float = Task.SrcScaleY
	Dim SrcRect1 As B4XRect = Task.SrcRect
	Dim SrcRectWidth As Int = SrcRect1.Width
	Dim SrcRectHeight As Int = SrcRect1.Height
	Dim SrcRectLeft As Int = SrcRect1.Left
	Dim SrcRectRight As Int = SrcRect1.Right
	Dim SrcRectTop As Int = SrcRect1.Top
	Dim SrcRectBottom As Int = SrcRect1.Bottom
	
	Dim TargetX As Int = Task.TargetX
	Dim TargetY As Int = Task.TargetY
	Dim Source As BitmapCreator = Task.Source
	Degrees = Degrees Mod 360
	If Degrees < 0 Then Degrees = 360 + Degrees
	Dim width As Int = SrcRectWidth * SrcScaleX
	Dim height As Int = SrcRectHeight * SrcScaleY
	Dim SinL As Float = SinD(Degrees)
	Dim CosL As Float = CosD(Degrees)
	Dim SSrcX, SSrcY, STargetX, STargetY As Int
	Dim x0, y0, x1, y1, x2, y2, x3, y3 As Float
	x0 = 0
	y0 = 0
	x1 = width * CosL
	y1 = width * SinL
	x2 = width * CosL - height * SinL
	y2 = width * SinL + height * CosL
	x3 = -height * SinL
	y3 =  height * CosL
	Dim TopX, TopY, RightX, RightY, BottomX, BottomY, LeftX, LeftY As Int
	If Degrees < 90 Then
		TopX = x0
		TopY = y0
		RightX = x1
		RightY = y1
		BottomX = x2
		BottomY = y2
		LeftX = x3
		LeftY = y3
		TargetX = Round(TargetX - (x3 + x1) / 2)
		TargetY = Round(TargetY - (y2 - y0) / 2)
	Else If Degrees < 180 Then
		TopX = x3
		TopY = y3
		RightX = x0
		RightY = y0
		BottomX = x1
		BottomY = y1
		LeftX = x2
		LeftY = y2
		TargetX = Round(TargetX + (-x2 + x0) / 2)
		TargetY = Round(TargetY - (y1  + y3) / 2)
	Else If Degrees < 270 Then
		TopX = x2
		TopY = y2
		RightX = x3
		RightY = y3
		BottomX = x0
		BottomY = y0
		LeftX = x1
		LeftY = y1
		TargetX = Round(TargetX - (x1 + x3) / 2)
		TargetY = Round(TargetY + (y0 - y2) / 2)
	Else
		TopX = x1
		TopY = y1
		RightX = x2
		RightY = y2
		BottomX = x3
		BottomY = y3
		LeftX = x0
		LeftY = y0
		TargetX = Round(TargetX - (-LeftX + RightX) / 2)
		TargetY = Round(TargetY - (BottomY + TopY) / 2)
	End If
	If Degrees Mod 90 = 0 Then
		Dim FixedTopY As Int = Max(-TargetY, TopY)
		Dim FixedBottomY As Int = Min(mHeight - TargetY, BottomY)
		Dim FixedLeftX As Int = Max(-TargetX, LeftX)
		Dim FixedRightX As Int = Min(mWidth - TargetX, RightX)
		Dim dx As Float = 1 / SrcScaleX
		Dim dy As Float = 1 / SrcScaleY
		Dim SrcX, SrcY, StartX, StartY As Float
		If Degrees = 0 Or Degrees = 180 Then
			If Degrees = 0 Then
				StartX = SrcRectLeft
				StartY = SrcRectTop
			Else
				StartX = SrcRectRight - dx
				StartY = SrcRectBottom - dy
				dx = -dx
				dy = -dy
			End If
			Dim OffsetX As Float = (FixedLeftX - LeftX) * dx
			SrcY = StartY + (FixedTopY - TopY) * dy
			For y = FixedTopY To FixedBottomY - 1
				SSrcY = SrcY 
				SrcX = StartX + OffsetX
				STargetY = TargetY + y
				For x = FixedLeftX To FixedRightX - 1
					SSrcX = SrcX 
					STargetX = TargetX + x
					#if B4i
					If Task.SkipBlending = False Then
						BlendPixel(Source, SSrcX, SSrcY, STargetX, STargetY)
					Else
						Dim SourceCP As Int = SSrcX * 4 + SSrcY * Source.mWidth * 4
						If Bit.FastArrayGetByte(Source.mBuffer, SSrcX * 4 + SSrcY * Source.mWidth * 4 + ai) > 0 Then
							Dim TargetCP As Int = STargetX * 4 + STargetY * mWidth * 4
							Bit.FastArraySetByte(mBuffer, TargetCP, Bit.FastArrayGetByte(Source.mBuffer, SourceCP))
							Bit.FastArraySetByte(mBuffer, TargetCP + 1, Bit.FastArrayGetByte(Source.mBuffer, SourceCP + 1))
							Bit.FastArraySetByte(mBuffer, TargetCP + 2, Bit.FastArrayGetByte(Source.mBuffer, SourceCP + 2))
							Bit.FastArraySetByte(mBuffer, TargetCP + 3, Bit.FastArrayGetByte(Source.mBuffer, SourceCP + 3))
						End If
					End If
					#else
					CopyPixelIgnoreSemiTransparent(Source, SSrcX, SSrcY, STargetX, STargetY , Task.SkipBlending)
					#end if
					SrcX = SrcX + dx
				Next
				SrcY = SrcY + dy
			Next
		Else If Degrees = 90 Or Degrees = 270 Then
			If Degrees = 90 Then
				StartY = SrcRectBottom - dy
				dy = -dy
				StartX = SrcRectLeft
			Else
				StartY = SrcRectTop
				StartX = SrcRectRight - dx
				dx = -dx
			End If
			SrcX = StartX + (FixedTopY - TopY) * dx
			Dim OffsetY As Float = dy * (FixedLeftX - LeftX)
			For y = FixedTopY To FixedBottomY - 1
				SrcY = StartY + OffsetY
				STargetY = y + TargetY
				SSrcX = SrcX
				For x = FixedLeftX To FixedRightX - 1
					SSrcY = SrcY
					STargetX = TargetX + x
					#if B4i
					If Task.SkipBlending = False Then
						BlendPixel(Source, SSrcX, SSrcY, STargetX, STargetY)
					Else
						Dim SourceCP As Int = SSrcX * 4 + SSrcY * Source.mWidth * 4
						If Bit.FastArrayGetByte(Source.mBuffer, SSrcX * 4 + SSrcY * Source.mWidth * 4 + ai) > 0 Then
							Dim TargetCP As Int = STargetX * 4 + STargetY * mWidth * 4
							Bit.FastArraySetByte(mBuffer, TargetCP, Bit.FastArrayGetByte(Source.mBuffer, SourceCP))
							Bit.FastArraySetByte(mBuffer, TargetCP + 1, Bit.FastArrayGetByte(Source.mBuffer, SourceCP + 1))
							Bit.FastArraySetByte(mBuffer, TargetCP + 2, Bit.FastArrayGetByte(Source.mBuffer, SourceCP + 2))
							Bit.FastArraySetByte(mBuffer, TargetCP + 3, Bit.FastArrayGetByte(Source.mBuffer, SourceCP + 3))
						End If
					End If
					#else
					CopyPixelIgnoreSemiTransparent(Source, SSrcX, SSrcY, STargetX, STargetY , Task.SkipBlending)
					#end if
					SrcY = SrcY + dy
				Next
				SrcX = SrcX + dx
			Next
		End If
		Return
	End If
	Dim c1 As Float = TanD(Degrees)
	Dim c2 As Float = CosL + c1 * SinL
	Dim ct2x As Float = 1 / c2 / SrcScaleX
	Dim ct2y As Float = c1 / c2 / SrcScaleY
	Dim SrcLeft As Int = SrcRectLeft
	Dim SrcTop As Int = SrcRectTop
	For y = Max(-TargetY, TopY) To Min(mHeight - 1 - TargetY, BottomY - 1)
		Dim XStart, XEnd As Int
		If y <= LeftY Then
			XStart = TopX + (y - TopY) * (LeftX - TopX) / (LeftY - TopY)
		Else
			XStart = LeftX + (y - LeftY) * (BottomX - LeftX) / (BottomY - LeftY)
		End If
		If y <= RightY Then
			XEnd = TopX + (y - TopY) * (RightX - TopX) / (RightY - TopY)
		Else
			XEnd = RightX + (y - RightY) * (BottomX - RightX) / (BottomY - RightY)
		End If
		XStart = Max(-TargetX, XStart)
		XEnd = Min(mWidth - TargetX, XEnd)
		Dim sx As Float = Min(Max((XStart + y * c1) / c2, 0), width - 1) / SrcScaleX
		Dim sy As Float = Min(Max((y - XStart * c1) / c2, 0), height - 1) / SrcScaleY
	
		Dim EndSX As Float = sx + ct2x * (XEnd - 1 - XStart)
		Do While (EndSX < 0 Or EndSX >= SrcRectRight - 0.1) And XEnd - 1 >= XStart
			XEnd = XEnd - 1
			EndSX = sx + ct2x * (XEnd - 1 - XStart)
		Loop
		Dim EndSY As Float = sy - ct2y * (XEnd - 1 - XStart)
		Do While (EndSY < 0 Or EndSY >= SrcRectBottom - 0.1) And XEnd - 1 >= XStart
			XEnd = XEnd - 1
			EndSY = sy - ct2y * (XEnd - 1 - XStart)
		Loop
		
		For x = XStart To XEnd - 1
			#if B4i
			SSrcX = SrcLeft + sx
			SSrcY = SrcTop + sy
			STargetX = TargetX + x
			STargetY = TargetY + y
			If Task.SkipBlending = False Then
				BlendPixel(Source, SSrcX, SSrcY, STargetX, STargetY)
			Else
				Dim SourceCP As Int = SSrcX * 4 + SSrcY * Source.mWidth * 4
				Dim a As Byte = Bit.FastArrayGetByte(Source.mBuffer, SSrcX * 4 + SSrcY * Source.mWidth * 4 + ai)
				If a > 0 Then
					Dim TargetCP As Int = STargetX * 4 + STargetY * mWidth * 4
					Bit.FastArraySetByte(mBuffer, TargetCP, Bit.FastArrayGetByte(Source.mBuffer, SourceCP))
					Bit.FastArraySetByte(mBuffer, TargetCP + 1, Bit.FastArrayGetByte(Source.mBuffer, SourceCP + 1))
					Bit.FastArraySetByte(mBuffer, TargetCP + 2, Bit.FastArrayGetByte(Source.mBuffer, SourceCP + 2))
					Bit.FastArraySetByte(mBuffer, TargetCP + 3, a)
				End If
			End If
			#else
			CopyPixelIgnoreSemiTransparent(Source, SrcLeft + sx, SrcTop + sy, TargetX + x, TargetY + y, Task.SkipBlending)
			#End If
			sx = sx + ct2x
			sy = sy - ct2y
		Next
	Next
	#if Stats
	Log($"DrawBitmapCreatorTransformed: ${DateTime.Now - Clock} ms"$)
	#End If
End Sub

Public Sub CopyPixelIgnoreSemiTransparent (Source As BitmapCreator, SrcX As Int, SrcY As Int, _
		 TargetX As Int, TargetY As Int, SkipBlending As Boolean) 'ignore
	If SkipBlending = False Then
		BlendPixel(Source, SrcX, SrcY, TargetX, TargetY)
		Return
	End If
	Dim SourceCP As Int = SrcX * 4 + SrcY * Source.mWidth * 4
	Dim a As Int = Source.mBuffer(SourceCP + ai)
	If a = 0 Then Return
	Dim TargetCP As Int = TargetX * 4 + TargetY * mWidth * 4
	mBuffer(TargetCP) = Source.mBuffer(SourceCP)
	mBuffer(TargetCP + 1) = Source.mBuffer(SourceCP + 1)
	mBuffer(TargetCP + 2) = Source.mBuffer(SourceCP + 2)
	mBuffer(TargetCP + 3) = Source.mBuffer(SourceCP + 3)
End Sub

'Copies a single pixel from Source to this BC. 
Public Sub CopyPixel (Source As BitmapCreator, SrcX As Int, SrcY As Int, TargetX As Int, TargetY As Int)
	Dim SourceCP As Int = SrcX * 4 + SrcY * Source.mWidth * 4
	Dim TargetCP As Int = TargetX * 4 + TargetY * mWidth * 4
#if B4i
	Bit.FastArraySetByte(mBuffer, TargetCP, Bit.FastArrayGetByte(Source.mBuffer, SourceCP))
	Bit.FastArraySetByte(mBuffer, TargetCP + 1, Bit.FastArrayGetByte(Source.mBuffer, SourceCP + 1))
	Bit.FastArraySetByte(mBuffer, TargetCP + 2, Bit.FastArrayGetByte(Source.mBuffer, SourceCP + 2))
	Bit.FastArraySetByte(mBuffer, TargetCP + 3, Bit.FastArrayGetByte(Source.mBuffer, SourceCP + 3))
#else
	mBuffer(TargetCP) = Source.mBuffer(SourceCP)
	mBuffer(TargetCP + 1) = Source.mBuffer(SourceCP + 1)
	mBuffer(TargetCP + 2) = Source.mBuffer(SourceCP + 2)
	mBuffer(TargetCP + 3) = Source.mBuffer(SourceCP + 3)
#End If
End Sub

'Copies a single pixel from Source to this BC. If the pixel is non-opaque then it will blend with the current pixel.
Public Sub BlendPixel (Source As BitmapCreator, SrcX As Int, SrcY As Int, TargetX As Int, TargetY As Int)
	Dim SrcPM As PremultipliedColor = Source.GetPremultipliedColor(SrcX, SrcY, tempPM2)
	If SrcPM.a <> UnsignedFF Then
		If SrcPM.a = 0 Then Return
		SrcPM.a = Bit.And(0xff, SrcPM.a)
		GetPremultipliedColor(TargetX, TargetY, tempPM)
		Dim SrcAF As Float = 1 - (SrcPM.a / 255)
		SrcPM.r =  Bit.And(0xff, SrcPM.r) + SrcAF *  Bit.And(0xff, tempPM.r)
		SrcPM.g =  Bit.And(0xff, SrcPM.g) + SrcAF *  Bit.And(0xff, tempPM.g)
		SrcPM.b =  Bit.And(0xff, SrcPM.b) + SrcAF *  Bit.And(0xff, tempPM.b)
		SrcPM.a =  Bit.And(0xff, SrcPM.a) + SrcAF *  Bit.And(0xff, tempPM.a)
	End If
	SetPremultipliedColor(TargetX, TargetY, SrcPM)
End Sub

Public Sub BlendPixel2 (SrcPM As PremultipliedColor, TargetX As Int, TargetY As Int)
	If SrcPM.a <> UnsignedFF Then
		If SrcPM.a = 0 Then Return
		tempPM2.a = Bit.And(0xff, SrcPM.a)
		GetPremultipliedColor(TargetX, TargetY, tempPM)
		Dim SrcAF As Float = 1 - (tempPM2.a / 255)
		tempPM2.r =  Bit.And(0xff, SrcPM.r) + SrcAF *  Bit.And(0xff, tempPM.r)
		tempPM2.g =  Bit.And(0xff, SrcPM.g) + SrcAF *  Bit.And(0xff, tempPM.g)
		tempPM2.b =  Bit.And(0xff, SrcPM.b) + SrcAF *  Bit.And(0xff, tempPM.b)
		tempPM2.a =  Bit.And(0xff, tempPM2.a) + SrcAF *  Bit.And(0xff, tempPM.a)
		SetPremultipliedColor(TargetX, TargetY, tempPM2)
	Else
		SetPremultipliedColor(TargetX, TargetY, SrcPM)
	End If
End Sub



#Region *************************  Gradients *****************************
'Fills the rectangle with a gradient. Skips blending.
'GradColors - An array of two or more colors that define the gradient.
'Rect - The region that will be filled.
'Orientation - One of the following: TL_BR, TOP_BOTTOM, TR_BL, LEFT_RIGHT, RIGHT_LEFT, BL_TR, BOTTOM_TOP, BR_TL and RECTANGLE.
Public Sub FillGradient (GradColors() As Int, Rect As B4XRect, Orientation As String)
	If Rect.Width <= 1 Or Rect.Height <= 1 Or Rect.Left >= mWidth Or Rect.Top >= mHeight Then Return
	If GradColors.Length = 1 Then
		FillRect(GradColors(0), Rect)
		Return
	End If
	Dim SinL As Float
	Dim StartY As Int = Max(Rect.Top, 0)
	Dim EndY As Int = Min(Rect.Bottom - 1, mHeight - 1)
	Dim ox, deltaox As Float
	Dim AxisLength As Int
	Dim TopBottom As Boolean
	Dim OppositeColors As Boolean
	Dim TR_BL As Boolean
	Dim Rectangle As Boolean
	Dim NumberOfParts As Int = GradColors.Length - 1
	Dim DFix(NumberOfParts + 1) As Float
	Dim DFixDivSinL(NumberOfParts + 1) As Float
	Select Orientation
		Case "TL_BR", "BR_TL", "TR_BL", "BL_TR"
			Dim alpha As Float = ATanD(Rect.Width / Rect.Height)
			SinL = SinD(alpha)
			AxisLength = Sqrt(Power(Rect.Width, 2) + Power(Rect.Height, 2))
			Dim TanL As Float = TanD(alpha)
			deltaox = 1 / TanL
			ox = (StartY - Rect.Top) / TanL
			OppositeColors = Orientation = "BR_TL" Or Orientation = "BL_TR"
			TR_BL = Orientation = "TR_BL" Or Orientation = "BL_TR"
		Case "LEFT_RIGHT", "RIGHT_LEFT"
			AxisLength = Rect.Width
			deltaox = 0
			SinL = 1
			OppositeColors = Orientation = "RIGHT_LEFT"
		Case "TOP_BOTTOM", "BOTTOM_TOP"
			AxisLength = Rect.Height
			SinL = 0
			ox = 0
			TopBottom = True
			OppositeColors = Orientation = "BOTTOM_TOP"
		Case "RECTANGLE"
			AxisLength = Max(Rect.Width, Rect.Height) / 2
			Rectangle = True
		Case Else
			Log("Invalid orientation: " & Orientation)
	End Select
	
	Dim RGBColor(GradColors.Length) As ARGBColor
	For i = 0 To GradColors.Length - 1
		Dim a As Int = i
		If OppositeColors Then a = GradColors.Length - 1 - i
		ColorToARGB(GradColors(a), RGBColor(i))
	Next
	
	For i = 0 To NumberOfParts
		DFix(i) = i / NumberOfParts * AxisLength
		DFixDivSinL(i) = DFix(i) / SinL
	Next
	Dim EndMax As Int = Min(mWidth - 1, Rect.Right - 1)
	Dim StartMin As Int = Max(Rect.Left, 0)
	Dim ClrFrom, ClrTo As ARGBColor
	Dim argb As ARGBColor
	Dim tStartX, tEndX, tStep = 1 As Int
	If Rectangle Then
		Dim cx As Int = Rect.CenterX
		Dim cy As Int = Rect.CenterY
		Dim width As Int = Rect.Width
		Dim height As Int = Rect.Height
		If (width Mod 2) = 0 Or (height Mod 2) = 0 Then AxisLength = AxisLength - 1
		Dim XFix = Min(1, width / height), YFix = Min(1, height / width) As Float
		For part = 0 To NumberOfParts - 1
			ClrFrom = RGBColor(part)
			ClrTo = RGBColor(part + 1)
			For r = part / NumberOfParts * AxisLength To (part + 1) / NumberOfParts * AxisLength
				Dim dd As Float = (r - DFix(part)) / (AxisLength / NumberOfParts)
				For x = -r To r
					argb.a = ClrFrom.a + dd * (ClrTo.a - ClrFrom.a)
					argb.r = ClrFrom.r + dd * (ClrTo.r - ClrFrom.r)
					argb.g = ClrFrom.g + dd * (ClrTo.g - ClrFrom.g)
					argb.b = ClrFrom.b + dd * (ClrTo.b - ClrFrom.b)
					ARGBToPremultipliedColor(argb, tempPM)
					SetPremultipliedColor(cx + x * XFix, cy - r * YFix, tempPM)
					SetPremultipliedColor(cx + x * XFix, cy + r * YFix, tempPM)
					SetPremultipliedColor(cx - r * XFix, cy + x * YFix, tempPM)
					SetPremultipliedColor(cx + r * XFix, cy + x * YFix, tempPM)
				Next
			Next
		Next
		Dim RectLeft = Rect.Left, RectRight = Rect.Right, RectTop = Rect.Top, RectBottom = Rect.Bottom As Int
		For x = RectLeft To RectRight - 1
			SetPremultipliedColor(x, RectTop, tempPM)
			SetPremultipliedColor(x, RectBottom - 1, tempPM)
		Next
		For y = RectTop To RectBottom - 1
			SetPremultipliedColor(RectLeft, y, tempPM)
			SetPremultipliedColor(RectRight - 1, y, tempPM)
		Next
	Else
		For y = StartY To EndY
			Dim EndX As Int = Rect.Left - ox - 1
			For part = 0 To NumberOfParts - 1
				ClrFrom = RGBColor(part)
				ClrTo = RGBColor(part + 1)
				If TopBottom Then
					Dim d As Float = y - Rect.Top
					Dim dd As Float = (d - DFix(part)) / (AxisLength / NumberOfParts)
					If dd < 0 Or dd > 1 Then
						Continue
					End If
					Dim tStartX As Int = StartMin
					Dim tEndX As Int = EndMax
				Else
					Dim StartX As Int = Max(StartMin, EndX + 1)
					Dim EndX As Int = Min(Rect.Left + DFixDivSinL(part + 1) - ox, EndMax)
					Dim d As Float = SinL * (StartX - Rect.Left + ox)
					Dim ddDelta As Float = SinL / (AxisLength / NumberOfParts)
					Dim dd As Float = (d - DFix(part)) / (AxisLength / NumberOfParts)
					If TR_BL Then
						tStartX = Rect.Right - 1 - (StartX - Rect.Left)
						tEndX = Rect.Right - 1 - (EndX - Rect.Left)
						tStep = -1
					Else
						tStartX = StartX
						tEndX = EndX
					End If
				End If
				For	x = tStartX To tEndX Step tStep
					argb.a = ClrFrom.a + dd * (ClrTo.a - ClrFrom.a)
					argb.r = ClrFrom.r + dd * (ClrTo.r - ClrFrom.r)
					argb.g = ClrFrom.g + dd * (ClrTo.g - ClrFrom.g)
					argb.b = ClrFrom.b + dd * (ClrTo.b - ClrFrom.b)
					SetARGB(x, y, argb)
					dd = dd + ddDelta
				Next
			Next
			ox = ox + deltaox
		Next
	End If
	
End Sub

'Fills the rectangle with a radial gradient. Skips blending.
Public Sub FillRadialGradient (GradColors() As Int, Rect As B4XRect)
	Dim RGBColor(GradColors.Length) As ARGBColor
	For i = 0 To GradColors.Length - 1
		ColorToARGB(GradColors(i), RGBColor(i))
	Next
	Dim MaxDistance As Int = Ceil(Sqrt(Power(Rect.Width / 2, 2) + Power(Rect.Height / 2, 2)))
	Dim DistanceToColor(MaxDistance) As ARGBColor
	Dim CurrentColorIndex As Int
	Dim LengthOfEachColor As Int = Ceil(MaxDistance / (GradColors.Length - 1))
	Dim c As Int
	For i = 0 To MaxDistance - 1
		Dim l As Float = CurrentColorIndex / LengthOfEachColor
		DistanceToColor(i).a = RGBColor(c).a + l * (RGBColor(c + 1).a - RGBColor(c).a)
		DistanceToColor(i).r = RGBColor(c).r + l * (RGBColor(c + 1).r - RGBColor(c).r)
		DistanceToColor(i).g = RGBColor(c).g + l * (RGBColor(c + 1).g - RGBColor(c).g)
		DistanceToColor(i).b = RGBColor(c).b + l * (RGBColor(c + 1).b - RGBColor(c).b)
		CurrentColorIndex = CurrentColorIndex + 1
		If CurrentColorIndex = LengthOfEachColor Then
			CurrentColorIndex = 0
			c = c + 1
		End If
	Next
	Dim cx As Int = Rect.CenterX
	Dim cy As Int = Rect.CenterY
	For x = Max(0, Rect.Left) To Min(mWidth - 1, Rect.Right - 1)
		Dim dx As Int = Power(x - cx, 2)
		For y = Max(0, Rect.Top) To Min(mHeight - 1, Rect.Bottom - 1)
			Dim distance As Int = Sqrt(dx + Power(y - cy, 2))
			SetARGB(x, y, DistanceToColor(distance))
		Next
	Next
End Sub
#End Region

#Region *****************  Native code ***********
#if Release
#if B4A
#if Java
public void DrawBitmapCreatorsAsync( final java.util.List<_drawtask> tasks, final String eventName, final B4AClass targetModule) {
	ba.submitRunnable(new Runnable() {
	public void run() {
		try {
			long clock = System.currentTimeMillis();
			for (_drawtask dt : tasks) {
				bitmapcreator target = dt.TargetBC == null ? bitmapcreator.this : dt.TargetBC;
				if (dt.Transform)
					target._drawbitmapcreatortransformed(dt);
				else if (dt.IsCompressedSource)
					target._drawcompressedbitmap((_compressedbc)dt.Source, dt.SrcRect, dt.TargetX, dt.TargetY);
				else
					target._drawbitmapcreator((bitmapcreator)dt.Source, dt.SrcRect, dt.TargetX, dt.TargetY, dt.SkipBlending);
			}
			//_clockforasync = System.currentTimeMillis();
			BA.handler.postAtFrontOfQueue(new Runnable() {
				public void run() {
					try {
						targetModule.getBA().raiseEvent2(null, false, eventName + "_bitmapready", false, _getbitmap());
						} catch (Exception e2) {
						e2.printStackTrace();
					}
				}
			});
			//BA.Log("" + (System.currentTimeMillis() - clock));
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	}, null, 0);
}
#end if
#else If B4J
#if Java
public void DrawBitmapCreatorsAsync( final java.util.List<_drawtask> tasks, final String eventName, final B4AClass targetModule) {
	ba.submitRunnable(new Runnable() {
	public void run() {
		try {
			long clock = System.currentTimeMillis();
			for (_drawtask dt : tasks) {
				bitmapcreator target = dt.TargetBC == null ? bitmapcreator.this : dt.TargetBC;
				if (dt.Transform)
					target._drawbitmapcreatortransformed(dt);
				else if (dt.IsCompressedSource)
					target._drawcompressedbitmap((_compressedbc)dt.Source, dt.SrcRect, dt.TargetX, dt.TargetY);
				else
					target._drawbitmapcreator((bitmapcreator)dt.Source, dt.SrcRect, dt.TargetX, dt.TargetY, dt.SkipBlending);
			}
			targetModule.getBA().raiseEventFromDifferentThread(null, null, 0, eventName + "_bitmapready", false, new Object[] {null});
			
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	}, null, 0);
}
#end if
#End If
#end if
#if RELEASE
#if OBJC
- (void) DrawBitmapCreatorsAsync:(NSArray*)tasks :(NSString*)eventName :(B4IClass*) targetModule{
	dispatch_async (dispatch_get_global_queue(DISPATCH_QUEUE_PRIORITY_DEFAULT, 0), ^{
		for (_drawtask* dt in tasks) {
			b4i_bitmapcreator* target = dt->_TargetBC == nil ? self : dt->_TargetBC;
			if (dt->_Transform)
				[target _drawbitmapcreatortransformed:dt];
			else if (dt->_IsCompressedSource)
				[target _drawcompressedbitmap:dt->_Source :dt->_SrcRect :dt->_TargetX :dt->_TargetY];
			else
				[target _drawbitmapcreator:dt->_Source :dt->_SrcRect :dt->_TargetX :dt->_TargetY :dt->_SkipBlending];
		}
		[targetModule.bi raiseEventFromDifferentThread:nil event:eventName params:@[[self _getbitmap]]];
	});
}
#end if
#end if
#if OBJC

- (void)UIImageToBuffer:(UIImage*) bmp :(NSData*)data :(int)width :(int)height :(BOOL)shouldClear{
	if (shouldClear)
		memset ([(NSMutableData*)data mutableBytes],0, data.length);
	CGColorSpaceRef colorSpace = CGColorSpaceCreateDeviceRGB();
	CGContextRef context = CGBitmapContextCreate([(NSMutableData*)data mutableBytes], 
			width, 
			height, 
			8, 
			4 * width, 
			colorSpace, 
			kCGImageAlphaPremultipliedLast); 
	if(!context) {
		NSLog(@"Failed To create bitmap context");
	}		
	CGColorSpaceRelease(colorSpace);
	CGImageRef imageRef = bmp.CGImage;
	CGRect Rect = CGRectMake(0, 0, width, height);
	CGContextDrawImage(context, Rect, imageRef);
	CGContextRelease(context);
}
- (UIImage*) bufferToUIImage:(NSData*)data :(int)width :(int)height {
	CGDataProviderRef provider = CGDataProviderCreateWithCFData( (CFDataRef) data );
	CGColorSpaceRef colorSpaceRef = CGColorSpaceCreateDeviceRGB();
	CGBitmapInfo bitmapInfo = kCGBitmapByteOrderDefault | kCGImageAlphaPremultipliedLast;
	CGColorRenderingIntent renderingIntent = kCGRenderingIntentDefault;
	
	CGImageRef iref = CGImageCreate(width, 
				height, 
				8, 
				32, 
				4 * width, 
				colorSpaceRef, 
				bitmapInfo, 
				provider,	
				NULL,	
				NO,
		renderingIntent);
	UIImage* image = [UIImage imageWithCGImage:iref scale:1 orientation:UIImageOrientationUp];
	CGColorSpaceRelease(colorSpaceRef);
	CGImageRelease(iref);
	CGDataProviderRelease(provider);
	return image;
}
#End If

#End Region